
//		Buffered file-write functions, for programs which would otherwise
//		call WriteFile() a bazillion times.


#ifndef BUFOUT_INCLUDED
#define BUFOUT_INCLUDED				1.0.6


#define BUFOUT_MIN_SIZE				4096
#define BUFOUT_MAX_SIZE				32768

#define BUFOUT_FILE_UTF8			0
#define BUFOUT_FILE_UTF16			1
#define BUFOUT_FILE_OEM				2
#define BUFOUT_FILE_ENCODING_MAX	2


#ifndef BOM
#define BOM							0xfeff
#endif	//	 BOM



const wchar_t BufOut_EncodingNames[BUFOUT_FILE_ENCODING_MAX + 2][8]	= { L"UTF-8", L"UTF-16", L"OEM", L"Unknown" };



struct BufOut_Buffer {
	wchar_t *data;
	char *xlatdata;
	unsigned int size;
	unsigned int ptr;
	unsigned int xlatsize;
	HANDLE handle;
	BOOL tostdout;
	unsigned int enc;
	unsigned int codepage;
	unsigned int errors;
	int lasterror;
	unsigned __int64 lines;
	unsigned __int64 bytes;
	unsigned __int64 chars;
	unsigned int writes;
};



unsigned int BufOut_CountEOLs( LPCTSTR String )
{
	if ( !String )
		return 0;

	const wchar_t *ch	= String;
	unsigned int k		= 0;


	while ( *ch ) {
		if ( *ch == '\r' ) {
			k++;

			if ( ch[1] == '\n' )
				ch++;
		}
		else if ( *ch == '\n' )
			k++;

		ch++;
	}

	return k;
}



unsigned int BufOut_CountChars( LPCTSTR String, BOOL IgnoreInitialBOM )
{
	if ( !String )
		return 0;

	const wchar_t *ch	= String;
	unsigned int k		= 0;


	while ( *ch ) {
		k++;

		if ( ( ch[0] & 0xfc00 ) == 0xd800 )
			if ( ( ch[1] & 0xfc00 ) == 0xdc00 )
				ch++;

		ch++;
	}

	if ( IgnoreInitialBOM && ( String[0] == BOM ) )
		k--;

	return k;
}



BufOut_Buffer *OpenBufferedFile( LPCTSTR Filename, unsigned int BufferSize, unsigned int Encoding, unsigned int CodePage )
{
	BufOut_Buffer *Buf		= NULL;
	HANDLE Handle			= NULL;
	int Error				= 0;


	if ( !Filename ) {
		Error = ERROR_INVALID_DATA;
		goto ErrorExit;
	}
	else if ( Filename[0] == '\0' ) {
		Error = ERROR_INVALID_DATA;
		goto ErrorExit;
	}


	if ( BufferSize < BUFOUT_MIN_SIZE )
		BufferSize = BUFOUT_MIN_SIZE;
	else if ( BufferSize > BUFOUT_MAX_SIZE )
		BufferSize = BUFOUT_MAX_SIZE;


	Buf = (BufOut_Buffer*) malloc( sizeof( BufOut_Buffer ) );
	if ( !Buf )
		goto NotEnoughMemory;


	if ( ( Encoding == BUFOUT_FILE_OEM ) && ( CodePage == CP_UTF8 ) )
		Encoding = BUFOUT_FILE_UTF8;

	memset( Buf, 0x00, sizeof( BufOut_Buffer ) );
	Buf->size			= BufferSize;
	Buf->enc			= Encoding;

	if ( Encoding == BUFOUT_FILE_OEM )
		Buf->codepage	= CodePage;
	else if ( Encoding == BUFOUT_FILE_UTF8 )
		Buf->codepage	= CP_UTF8;

	
	Buf->data		= (wchar_t*) malloc( BufferSize * sizeof( wchar_t ) );
	if ( Buf->data == NULL ) 
		goto NotEnoughMemory;

	Buf->data[0] = '\0';

	if ( Encoding == BUFOUT_FILE_UTF8 )
		Buf->xlatsize = BufferSize * 3;
	else if ( Encoding == BUFOUT_FILE_OEM )
		Buf->xlatsize = BufferSize * 2;

	if ( Buf->xlatsize ) {
		Buf->xlatdata = (char*) malloc( Buf->xlatsize );
		if ( Buf->xlatdata == NULL )
			goto NotEnoughMemory;

		Buf->xlatdata[0] = '\0';
	}


	if ( wcscmp( Filename, L"-" ) )
		Handle = CreateFile( Filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
	else {
		Handle = GetStdHandle( STD_OUTPUT_HANDLE );
		Buf->tostdout = TRUE;
	}

	if ( !Handle || ( Handle == INVALID_HANDLE_VALUE ) ) {
		Error = GetLastError();
		goto ErrorExit;
	}

	Buf->handle = Handle;
	return Buf;


NotEnoughMemory:

	Error = ERROR_NOT_ENOUGH_MEMORY;
	//	fall through to...

ErrorExit:

	if ( Buf ) {
		free( Buf->data );
		free( Buf->xlatdata );
		free( Buf );
	}

	SetLastError( Error );
	return NULL;
}



int FlushOutBuffer( BufOut_Buffer *Buf )
{
	if ( !Buf )
		return -1;

	if ( Buf->ptr == 0 )
		return 0;


	DWORD BytesToWrite		= 0;
	DWORD BytesWritten		= 0;
	int rv					= 0;


	switch ( Buf->enc ) {

	case BUFOUT_FILE_UTF16 :

		rv = WriteFile( Buf->handle, Buf->data, (DWORD) Buf->ptr * sizeof( wchar_t ), &BytesWritten, NULL );
		break;

	case BUFOUT_FILE_OEM :
	case BUFOUT_FILE_UTF8 :
	default :

		BytesToWrite = WideCharToMultiByte( Buf->codepage, 0, Buf->data, -1, Buf->xlatdata, Buf->xlatsize, NULL, NULL );
		rv = WriteFile( Buf->handle, Buf->xlatdata, BytesToWrite - 1, &BytesWritten, NULL );
		break;
	}


	if ( !rv ) {
		rv = GetLastError();
		Buf->errors += 1;
		Buf->lasterror = rv;
		return rv;
	}


	Buf->lines += BufOut_CountEOLs( Buf->data );
	Buf->chars += BufOut_CountChars( Buf->data, ( Buf->bytes == 0 ) );
	Buf->bytes += BytesWritten;
	Buf->writes += 1;
	Buf->ptr = 0;
	Buf->data[0] = '\0';
	return 0;
}



int BufferedOutput( BufOut_Buffer *Buf, LPCTSTR String )
{
	if ( !Buf || !String )
		return -1;


	size_t len		= wcslen( String );
	int rv			= 0;


	if ( ( Buf->ptr + len ) >= Buf->size )
		if ( rv = FlushOutBuffer( Buf ) )
			return rv;


	if ( len >= Buf->size ) {		//	this bit looks jolly clever, but has never been tested!

		unsigned int ts		= ( Buf->size / 8 ) * 7;
		unsigned int n		= ts;

		while ( wcslen( String ) >= Buf->size ) {
			n = ts;
			if ( ( Buf->data[n - 1] & 0xfc00 ) == 0xd800 )		//	don't split a surrogate pair!
				if ( ( Buf->data[n] & 0xfc00 ) == 0xdc00 )
					n--;

			wcsncpy_s( Buf->data, Buf->size, String, n );
			Buf->data[n] = '\0';
			Buf->ptr = n;

			if ( rv = FlushOutBuffer( Buf ) )
				return rv;

			String += n;
		}
	}


	wcscat_s( Buf->data, Buf->size, String );
	Buf->ptr = (unsigned int) wcslen( Buf->data );
	return 0;
}



int BufferedOutput( BufOut_Buffer *Buf, const wchar_t ch )
{
	if ( !Buf )
		return -1;

	if ( !ch )
		return 0;


	int rv			= 0;


	if ( ( Buf->ptr + 1 ) >= Buf->size ) {
		if ( rv = FlushOutBuffer( Buf ) )
			return rv;
	}

	Buf->data[Buf->ptr] = ch;
	Buf->ptr += 1;
	Buf->data[Buf->ptr] = '\0';
	return 0;
}



#ifdef BUFOUT_PRINTF

int BufOutPrintf( BufOut_Buffer *Buf, LPCTSTR Format, ... )
{
	const int buflen			= BUFOUT_MIN_SIZE / 2;
	wchar_t temp[buflen]		= L"";
	va_list Args;


	va_start( Args, Format );
	vswprintf( temp, buflen, Format, Args );
	va_end( Args );

	return BufferedOutput( Buf, temp );
}

#endif	// BUFOUT_PRINTF



int CloseBufferedFile( BufOut_Buffer *Buf )
{
	if ( !Buf )
		return -1;


	int rv			= 0;


	rv = FlushOutBuffer( Buf );

	if ( Buf->tostdout == FALSE )
		CloseHandle( Buf->handle );

	free( Buf->xlatdata );
	free( Buf->data );
	free( Buf );
	return rv;
}



LPCTSTR BufOut_EncodingName( unsigned int Enc )
{
	if ( Enc <= BUFOUT_FILE_ENCODING_MAX )
		return BufOut_EncodingNames[Enc];
	else
		return BufOut_EncodingNames[BUFOUT_FILE_ENCODING_MAX + 1];
}


#endif	// BUFOUT_INCLUDED


/*

	1.0.6	2023-03-15
			Filename arg to OpenBufferedFile() is declared as a constant
			string.

	1.0.5	2021-10-23
			Reduced the size of the UTF-8 translation buffer.

	1.0.4	2021-07-11
			A filename of "-" uses stdout.

	1.0.3	2020-11-05
			Now maintains a count of characters output, along with bytes,
			lines, etc.

	1.0.2	2020-09-22
			The line count routine now recognizes CR and CR/LF as well
			as LF.

	1.0.1	2020-06-09

*/
